/*  VER 158  TAB P   $Id: telnet.c,v 1.9.2.4 2003/01/21 09:47:25 egil Exp $
 *
 *  implement a very simple telnet protocol
 *
 *  copyright 1996, 1997 Egil Kvaleberg, egil@kvaleberg.no
 *  the GNU General Public License applies
 *
 *  $Log: telnet.c,v $
 *  Revision 1.9.2.4  2003/01/21 09:47:25  egil
 *  Renamed MAXHEADERSIZE to MAX_HEADER_SIZE due to name collision
 *
 *  Revision 1.9.2.3  2002/10/01 08:54:59  egil
 *  Changed sprintf to snprintf
 *
 *  Revision 1.9.2.2  2002/09/21 17:20:36  egil
 *  A long range of patches incorporated..
 *
 *  Revision 1.9.2.1  2002/01/29 06:44:48  egil
 *  Changing from xmalloc, xrealloc, xstrcpy to
 *  malloc_perfect, realloc_perfect and strdup_perfect
 *
 *  Revision 1.9  1999/03/31 05:53:46  src
 *  Seperated MAXHEADERSIZE from NNTP_STRLEN
 *
 *  Revision 1.8  1999/03/31 03:38:29  src
 *  Moved errno.h to common.h
 *
 *  Revision 1.7  1998/09/09 07:32:14  src
 *  Version 1.1
 *
 *  Revision 1.6  1998/09/03 02:49:31  src
 *  Fixed stuff detected by -Wall
 *
 *  Revision 1.5  1998/07/12 09:39:30  src
 *  newsx version 1.0
 */

#include "common.h"
#include "proto.h"
#include "nntp.h"
#include "news.h" /* MAX_HEADER_SIZE */

#include <signal.h>
#include <setjmp.h>

#include <netdb.h>
#include <sys/socket.h>

#define TELOPTS
#define TELCMDS
#include <arpa/telnet.h>

#ifndef IAC
#define IAC  255
#define DONT 254 
#define DO   253
#define WONT 252
#define WILL 251
#define TELOPT_BINARY 0
#define TELOPT_ECHO 1
#define TELOPT_SGA 3 
#define TELOPT_LINEMODE 34
#endif  

/*
 *  write line with telnet protocol
 *  suppress echo
 *  return false if a fatal error occurred
 */
int 
write_telnet_line(char *line, int len, SOCKET_D *sock)
{
    int n;
    char c1,c2;
    char echo[MAX_HEADER_SIZE+3];
    int wrong_echo = 0;

    /* BUG: should escape IACs */
    /* BUG: should we use flag MSG_DONTWAIT? */
    if (send(sock->w_fd,line,len,0) != len
     || send(sock->w_fd,"\r",1,0) != 1) { /* using CR as newline for telnet */
	log_msg(L_ERRno,"write error on telnet socket");
	return 0;
    }
    if (sock->is_echo) {
      again:
	/* await echo */
	echo[MAX_HEADER_SIZE+2] = '\0';
	if (read_telnet(echo,MAX_HEADER_SIZE+2,sock) < 0) return 0;
	for (n=0; n<len; ++n) {
	    c1 = line[n];  
	    c2 = echo[n];  
	    if ((c1=='\r' || c1=='\n')
	     && (c2=='\r' || c2=='\n' || !c2)) break;

	    /* NOTE: echo may be filtered through 7bit mask */
	    if ((line[n] & 0x7f) != (echo[n] & 0x7f)) {
		log_msg(L_DEBUG,"wrong echo \"%s\"",echo);
		if (++wrong_echo >= 100) {
		    /* prevent us hanging here forever... */
		    log_msg(L_ERR,"persistent wrong echo: \"%s\"", echo);
		    exit_cleanup(11);
		}
		goto again;
	    }
	}
    }
    return 1;
}

/*
 *  read line with telnet protocol
 *  NOTE: this is really a primitive implementation!
 */
static void 
send_iac3(SOCKET_D *sock,char how,char what)
{
    char sbuf[3];
    sbuf[0] = IAC;
    sbuf[1] = how;
    sbuf[2] = what;
    send(sock->w_fd,sbuf,3,0);
}

/*
 *  return string for telnet command
 */
static char *
str_telcmd(int cmd)
{
    static char str[10];
#ifdef TELCMD
    if (TELCMD_OK(cmd)) {
	return TELCMD(cmd);
    }
#endif
    snprintf(str,sizeof(str),"%d",cmd);
    return str;
}

/*
 *  return string for telnet option
 */
static char *
str_telopt(int opt)
{
    static char str[10];
#ifdef TELOPT
    if (TELOPT_OK(opt)) {
	return TELOPT(opt);
    }
#endif
    snprintf(str,sizeof(str),"%d",opt);
    return str;
}

/*
 *  read line with telnet protocol
 *  return string length, negative if a fatal error occurred
 *  tries to make sense out of various CR-LF peculiarities
 *  NOTE: this is really a primitive implementation!
 */
int 
read_telnet(char *line, int size, SOCKET_D *sock)
{
    int ix;
    int state;
    int c;
    int d;
    int is_cr = 0;
    static unsigned char *buf = 0;
    static int buf_n = 0;
    static int buf_m = 0;
    static int not_first_time = 0;

    if (!buf) buf = malloc_perfect(MAX_HEADER_SIZE);

    if (!not_first_time) {
	not_first_time = 1;
	/* IAC DO TELOPT_BINARY */
	/* IAC DO TELOPT_SGA = suppress goahead */
	/* IAC DONT TELOPT_ECHO (which is default) */
	/*
	log_msg(L_DEBUGMORE,"telnet send: IAC DO BINARY, IAC DO SGA");
	send_iac3(sock,DO,  TELOPT_BINARY);
	send_iac3(sock,DO,  TELOPT_SGA);
	 */
	/*
	send_iac3(sock,DONT,TELOPT_ECHO);
	send_iac3(sock,DONT,TELOPT_LINEMODE);
	 */
    }
    ix = 0;
    state = 0;
    while (ix < size) {
	/* read a packet */
	while (buf_n >= buf_m) {
	    /* need to refill local buffer */
	    buf_n = 0;
	    buf_m = recv(sock->r_fd,buf,MAX_HEADER_SIZE,0);
	    if (buf_m < 0) {
		log_msg(L_ERRno,"read error on telnet socket");
		return -1;
	    }
	    if (buf_m == 0) {
		/* [ver 0.10] there is nothing more... */
		line[ix] = '\0';
		return ix;
	    }
	}
	c = buf[buf_n++];

	switch (state) {
	default:
	    if (c == IAC) {
		state = IAC;
	    } else if (c == '\0') {
		/* ignore */
#if 0
	    } else if (ix == 0 && c == '\r') {
		/* ignore */
#endif
	    } else {
	      got_ch:
#if 1
		if (is_cr && c != '\n') {
		    /* carriage return by itself */
		    --buf_n; /* push back character */
		    c = '\n'; /* and fake a newline too */
		}      
#endif
		line[ix++] = c;
		if (ix >= size) return ix;
		is_cr = (c == '\r');
		line[ix] = '\0';

		/* NOTE: update for every new character */
		if (sock->f_chat)
		    chat_update(line,sock);

		if (c == '\n') {
		    return ix; /* end of line */
		}
	    }
	    break;
	case IAC:
	    switch (c) {
	    case IAC:
		state = 0;
		goto got_ch;
	    case DO:
	    case DONT:
	    case WILL:
	    case WONT:
		state = c;
		break;
	    case GA: /* go-ahead */
	    default:
		/* ignore two characters... */
		log_msg(L_DEBUGMORE,"telnet ignore: IAC %d",
						str_telcmd(c));
		state = 0;
		break;
	    }
	    break;
	case DO:
	    switch (c) {
	    case TELOPT_SGA:
	    case TELOPT_BINARY:
		send_iac3(sock,d=WILL,c);
		break;
	    case TELOPT_ECHO:
	    default:
		send_iac3(sock,d=WONT,c);
		break;
	    }
	    log_msg(L_DEBUGMORE,"telnet %s: IAC DO %s",
				     d==WILL ? "will":"wont",
						str_telopt(c));
	    state = 0;
	    break;
	case DONT:
	    log_msg(L_DEBUGMORE,"telnet ignore: IAC DONT %s",
						str_telopt(c));
	    state = 0;
	    break;
	case WILL:
	case WONT:
	    log_msg(L_DEBUGMORE,"telnet ignore: IAC %s %s",
						str_telcmd(state),
						str_telopt(c));
	    /* ignore three characters... */
	    state = 0;
	    break;
	}
    }
    return ix;
}
